import os
import math
from pylab import cm
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from . import *
import inspect
from inspect import isfunction
from typing import Callable, Any


class Drawer:
    algorithm_func_table = {}
    valid_parameters = {
        "has_aver",
        "log_times",
        "mat_size",
        "mtx",
        "coo_shape",
        "coo_data",
        "coo_rows",
        "coo_cols",
        "mat",
        "div",
        "row_size",
        "col_size",
        "row_block_sz",
        "col_block_sz",
        "real_max_data",
        "real_min_data",
    }
    color_theme = {
        "Default": colors.LinearSegmentedColormap.from_list("Default", ["blue", "white", "red"]),
        "SuiteSparse": colors.LinearSegmentedColormap.from_list("SuiteSparse", ["black", "blue", "cyan", "white", "green", "orange", "yellow"])
    }

    @classmethod
    def algorithmWrapper(cls):
        """
        Drawer 的算法装饰器, 已内置 aver, log, real, abs四种算法,
        您通过此装饰器可以轻松自定义算法, 支持直接获取Drawer对象支持的参数, 主要包含如下内容,
        (如自定义算法需要其他参数输入, 参数名需要以extern_开头, 并需要以extern_xxx=bala的形式赋值)
        has_aver: 是否有取平均值选项 => div是否可用

        log_times: 外部设定的取log的次数

        mat_size: 矩阵行列值较大的属性被分的块数

        mtx: 文件的scipy.sparse.coo_matrix对象, 未做任何更改

        coo_shape: mtx的尺寸

        coo_data: 矩阵的非零元值 (None if using streaming method)

        coo_rows: 矩阵的非零元素行索引映射到mat的行值 (None if using streaming method)

        coo_cols: 矩阵的非零元素列索引映射到mat的列值 (None if using streaming method)

        real_max_value: 矩阵非零元素最大值

        real_min_value: 矩阵非零元素最小值

        mat: 被初始化好的二维画布对象, 类型为numpy.array

        div: 子矩阵非零元数, 只有当has_aver为True时才会有效

        row_size: mat的行数

        col_size: mat的列数

        row_block_sz: 划分的子矩阵的行数

        col_block_sz: 划分的子矩阵的列数
        """

        def wrapper(func: Callable[[Any], float]):
            if not isfunction(func):
                raise TypeError("Algorithm Wrapper should wrap a function!")
            func_analyser = inspect.signature(func)
            func_name = func.__name__.strip("_")
            if func_name in Drawer.algorithm_func_table:
                raise Exception(f"Algorithm '{func_name}' already exists!")
            for arg in func_analyser.parameters.values():
                if arg.name not in Drawer.valid_parameters and not arg.name.startswith(
                    "extern_"
                ):
                    console.print(erro_string, f'Not support: "{arg.name}"')
                    raise Exception(
                        "Arg must in supported args or startswith 'extern_'"
                    )
            Drawer.algorithm_func_table[func_name] = {
                "func": func,
                "analyser": func_analyser,
            }

        return wrapper

    def __init__(
        self,
        filepath: str,
        has_aver: bool,
        force_update: bool = False,
        set_log_times: int = 2,
        set_mat_size: int = 200,
        set_block_size: int = -1,
        set_tick_step: int = -1,
        set_color_theme: str = "Default",
        img_format: str = "svg",
        font_color: str = "black",
        color_bar: bool = True,
        show_in_console: bool = False,
        parallel: bool = False,
    ):
        """
        初始化Drawer对象
        :param filepath: 矩阵文件路径
        :param has_aver: 是否有取平均值选项
        :param force_update: 是否强制更新图像
        :param set_log_times: 取log的次数
        :param set_mat_size: 缩略图尺寸
        :param set_block_size: 设置块大小（此参数设置后将覆盖set_mat_size）
        :param img_format: 图像格式
        :param font_color: 字体颜色
        :param show_in_console: 是否在控制台显示图像, 如果为True则不会保存图像
        """
        self.img_format = img_format
        self.font_color = font_color
        self.filepath = filepath
        self.force_update = force_update
        self.has_aver = has_aver
        self.img_path = filepath.split(".")
        self.img_path = ".".join(self.img_path[:-1]) + "_{}" + f".{img_format}"
        self.log_times = set_log_times
        self.color_map = Drawer.color_theme[set_color_theme]
        self.color_bar = color_bar
        self.show_in_console = show_in_console
        if self.show_in_console:
            console.print(info_string, "已开启控制台显示图像")
        elif not parallel:
            console.print(info_string, f'路径模板: "{self.img_path}"')

        self.mat_size = set_mat_size
        self.block_sz = set_block_size
        self.tick_step = set_tick_step

        (
            self.mtx,
            self.coo_shape,
            self.coo_data,
            self.coo_rows,
            self.coo_cols,
            self.raw_mat,
            self.mat,
            self.div,
            self.row_size,
            self.col_size,
            self.row_block_sz,
            self.col_block_sz,
            self.x_ticks,
            self.y_ticks,
        ) = (
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
        )
        self.absVal = 1
        self.parallel = parallel

    def loadMtx(self):
        if not self.parallel:
            status.update("加载矩阵")
        try:
            self.mtx = requirePackage("scipy.sparse", "coo_matrix")(
                requirePackage("scipy.io", "mmread")(self.filepath)
            )
        except ValueError:
            console.print(
                erro_string, "读取矩阵失败: 可能此文件格式不符合标准Matrix Market格式, 或位置值不以1作为第一个元素"
            )
            raise ValueError("读取矩阵失败: 可能此文件格式不符合标准Matrix Market格式, 或位置值不以1作为第一个元素")
        self.coo_shape = self.mtx.shape
        if self.block_sz > 0:
            self.row_size = int(math.ceil(self.coo_shape[0] / self.block_sz))
            self.col_size = int(math.ceil(self.coo_shape[1] / self.block_sz))
            self.row_block_sz = self.col_block_sz = self.block_sz
            self.y_ticks = np.linspace(0, self.row_size, 4)
            self.x_ticks = np.linspace(0, self.col_size, 4)
        else:
            self.mat_size = min(max(self.mtx.shape), self.mat_size)
            if self.coo_shape[0] >= self.coo_shape[1]:
                rate = self.coo_shape[1] / self.coo_shape[0]
                self.row_size = self.mat_size
                self.row_block_sz = self.coo_shape[0] / self.row_size
                self.col_size = int(math.ceil(self.mat_size * rate))
                self.col_block_sz = self.coo_shape[1] / self.col_size
                self.y_ticks = np.linspace(0, self.row_size, 4)
                self.x_ticks = np.linspace(0, self.col_size, max(my_round(4 * rate), 2))
            else:
                rate = self.coo_shape[0] / self.coo_shape[1]
                self.col_size = self.mat_size
                self.col_block_sz = self.coo_shape[1] / self.col_size
                self.row_size = int(math.ceil(self.mat_size * rate))
                self.row_block_sz = self.coo_shape[0] / self.row_size
                self.x_ticks = np.linspace(0, self.col_size, 4)
                self.y_ticks = np.linspace(0, self.row_size, max(my_round(4 * rate), 2))

        self.coo_data = self.mtx.data
        self.real_max_data = max(self.coo_data)
        self.real_min_data = min(self.coo_data)
        self.coo_rows = np.floor(self.mtx.row / self.row_block_sz).astype(np.int32)
        self.coo_cols = np.floor(self.mtx.col / self.col_block_sz).astype(np.int32)
        self.raw_mat = np.zeros((self.row_size, self.col_size), dtype=float)
        if not self.parallel:
            status.update("生成画布")
        if self.has_aver:
            self.div = np.ones((self.row_size, self.col_size), dtype=float)
            for i in zip(self.coo_data, self.coo_rows, self.coo_cols):
                self.raw_mat[i[1:]] += i[0]
                self.div[i[1:]] += 1
            self.div[self.div > 1] -= 1
        else:
            for i in zip(self.coo_data, self.coo_rows, self.coo_cols):
                self.raw_mat[i[1:]] += i[0]

    def loadMtx_streaming(self):
        if not self.parallel:
            status.update("加载矩阵并更新画布")

        from .MtxReader import mat_gen

        info = mat_gen(
            self.filepath,
            int(self.block_sz),
            int(self.mat_size),
            1 if self.has_aver else 0,
        )
        """
        rows, cols -> coo shape
        trows, tcols -> mat shape
        raw_mat -> raw_mat
        div_mat -> div
        """
        self.row_size = info["trows"]
        self.col_size = info["tcols"]
        self.row_block_sz = info['rows'] / self.row_size if self.block_sz == -1 else self.block_sz
        self.col_block_sz = info['cols'] / self.col_size if self.block_sz == -1 else self.block_sz
        self.coo_shape = (info["rows"], info["cols"])
        self.raw_mat = info["raw_mat"]
        if self.has_aver:
            self.div = info["div_mat"]

        self.real_max_data = info["real_max_value"]
        self.real_min_data = info["real_min_value"]
        self.x_ticks = (
            np.linspace(0, self.col_size, 4, True, dtype=np.int32)
            if self.tick_step == -1
            else np.array(
                [i for i in range(0, self.col_size, self.tick_step)] + [self.col_size],
                dtype=np.int32,
            )
        )
        self.y_ticks = (
            np.linspace(0, self.row_size, 4, True, dtype=np.int32)
            if self.tick_step == -1
            else np.array(
                [i for i in range(0, self.row_size, self.tick_step)] + [self.row_size],
                dtype=np.int32,
            )
        )

    def call(self, func_name: str, **kw_extern_args):
        from multiprocessing import Process

        if not self.parallel:
            status(f"正在执行: {func_name}").start()
        func_name = func_name.strip("_")
        if func_name not in Drawer.algorithm_func_table:
            raise KeyError(f"Algorithm '{func_name}' not registered!")
        analyser = Drawer.algorithm_func_table[func_name]["analyser"]
        algorithm = Drawer.algorithm_func_table[func_name]["func"]
        args_body = {}
        if self.raw_mat is None:
            self.loadMtx_streaming()
        # status(f"正在执行: {func_name}")
        self.mat = self.raw_mat.copy()
        for arg in analyser.parameters.values():
            if arg.name in Drawer.valid_parameters:
                args_body[arg.name] = getattr(self, arg.name)
        args_body.update(kw_extern_args)
        if os.path.exists(self.img_path.format(func_name)) and not self.force_update:
            return
        self.absVal = algorithm(**args_body)
        if not self.parallel:
            self.draw(func_name)
        else:
            process = Process(target=self.draw, args=(func_name,))
            process.start()
            process.join()  # 等待绘图完成

    def draw(self, suffix: str):
        if self.mat is None:
            return
        if not self.parallel:
            console.print(info_string, "absVal =", self.absVal)

        fig = plt.figure()
        plt.imshow(
            self.mat,
            origin="upper",
            cmap=self.color_map,
            norm=cm.colors.Normalize(vmin=-self.absVal, vmax=self.absVal),
            extent=[0, self.col_size, self.row_size, 0],
        )
        ax = plt.gca()
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position("top")
        plt.xticks(ticks=self.x_ticks, labels=np.append(self.x_ticks[:-1] * self.col_block_sz, self.coo_shape[1]).astype(np.int32))
        plt.yticks(ticks=self.y_ticks, labels=np.append(self.y_ticks[:-1] * self.row_block_sz, self.coo_shape[0]).astype(np.int32))
        if self.tick_step != -1:
            plt.grid(color="black", linestyle="-", linewidth=1)
        plt.tick_params(axis="x", colors=self.font_color)
        plt.tick_params(axis="y", colors=self.font_color)
        if self.color_bar:
            plt.colorbar().ax.tick_params(labelcolor=self.font_color)
        else:
            plt.colorbar().remove()
        # 设置背景透明
        fig.patch.set_alpha(0)
        fig.tight_layout()
        plt.subplots_adjust(left=0.00, wspace=0.0)

        if not self.show_in_console:
            fig.savefig(
                self.img_path.format(suffix),
                format=self.img_format,
                bbox_inches='tight',  # 确保边界紧凑
                dpi=300,
            )
        else:
            requirePackage(
                "QuickStart_Rhy.ImageTools.ImagePreview",
                "image_preview",
                "QuickStart_Rhy",
            )(fig)
        plt.close(fig)


@Drawer.algorithmWrapper()
def aver(mat, div):
    mat /= div
    return max(abs(max([max(i) for i in mat])), abs(min([min(i) for i in mat])))


@Drawer.algorithmWrapper()
def _abs(mat):
    mat[mat > 0] = 1
    mat[mat < 0] = -1
    return 1


@Drawer.algorithmWrapper()
def real(real_max_data, real_min_data) -> float:
    return max(abs(real_max_data), abs(real_min_data))


@Drawer.algorithmWrapper()
def log(mat: np.ndarray, log_times: int) -> float:
    for _ in range(log_times):
        mat[mat > 0] = np.log(mat[mat > 0])
        mat[mat < 0] = -np.log(-mat[mat < 0])
    return max(abs(max([max(i) for i in mat])), abs(min([min(i) for i in mat])))



import numpy as cnp
cimport numpy as cnp
from libc.stdlib cimport free

cdef extern from "numpy/arrayobject.h":
    ctypedef cnp.npy_intp npy_intp
    cnp.ndarray PyArray_SimpleNew(int nd, npy_intp* dims, int typenum)
    cnp.ndarray PyArray_SimpleNewFromData(int nd, npy_intp* dims, int typenum, void* data)

cdef int _numpy_initialized = cnp.import_array()

cdef extern from "_MatGen.c":
    ctypedef struct ThumbnailMatrix:
        int rows
        int cols  # original size (rows, cols)
        int trows
        int tcols # thumbnail size (trows, tcols)
        double* raw_mat # data of the matrix (trows, tcols)
        double real_max_value
        double real_min_value
        double* div_mat    # division matrix (trows, tcols)
    ThumbnailMatrix mat_gen_impl(const char* file_path, int block_sz, int mat_sz, int using_div)

def mat_gen(str file_path, int block_sz, int mat_sz, int using_div):
    """
    Read matrix from file_path and return a dense matrix.
    """
    cdef ThumbnailMatrix mat = mat_gen_impl(file_path.encode('utf-8'), block_sz, mat_sz, using_div)
    cdef cnp.ndarray[cnp.double_t, ndim=2] raw_mat
    cdef cnp.ndarray[cnp.double_t, ndim=2] div_mat
    if mat.raw_mat == NULL:
        raise ValueError("Cannot read matrix from file: {}".format(file_path))
    cdef npy_intp dims[2]
    dims[0] = mat.trows
    dims[1] = mat.tcols
    raw_mat = cnp.PyArray_SimpleNewFromData(2, dims, cnp.NPY_DOUBLE, <void*>mat.raw_mat).copy()
    free(mat.raw_mat)
    if using_div:
        div_mat = cnp.PyArray_SimpleNewFromData(2, dims, cnp.NPY_DOUBLE, <void*>mat.div_mat).copy()
        free(mat.div_mat)
    else:
        div_mat = None
    
    return {
        "rows": mat.rows,
        "cols": mat.cols,
        "trows": mat.trows,
        "tcols": mat.tcols,
        "raw_mat": raw_mat,
        "div_mat": div_mat,
        "real_max_value": mat.real_max_value,
        "real_min_value": mat.real_min_value
    }



from QuickProject import QproDefaultConsole as console
from QuickProject import QproInfoString as info_string
from QuickProject import QproWarnString as warn_string
from QuickProject import QproErrorString as erro_string
from QuickProject import QproDefaultStatus as status
from QuickProject import _ask, user_lang, external_exec, user_pip
from decimal import getcontext, Decimal, ROUND_HALF_EVEN
import sys

name = 'MtxDrawer'
getcontext().rounding = ROUND_HALF_EVEN


def my_round(x: float) -> int:
    return int(Decimal(x).quantize(Decimal('1')))


def requirePackage(
    pname: str,
    module: str = "",
    real_name: str = "",
    not_exit: bool = True,
    not_ask: bool = False,
    set_pip: str = user_pip,
):
    """
    获取本机上的python第三方库，如没有则询问安装

    :param not_ask: 不询问，无依赖项则报错
    :param set_pip: 设置pip路径
    :param pname: 库名
    :param module: 待引入的模块名，可缺省
    :param real_name: 用于 pip3 install 的名字
    :param not_exit: 安装后不退出
    :return: 库或模块的地址
    """
    local_scope = {}
    try:
        exec((f"from {pname} import {module}" if module else f"import {pname}"), globals(), local_scope)
    except (ModuleNotFoundError, ImportError):
        if not_ask:
            return None
        if _ask(
            {
                "type": "confirm",
                "message": f"""{name} require {pname + (' -> ' + module if module else '')}, confirm to install?
  {name} 依赖 {pname + (' -> ' + module if module else '')}, 是否确认安装?""",
                "default": True,
            }
        ):
            with status("Installing..." if user_lang != "zh" else "正在安装..."):
                external_exec(
                    f"{set_pip} install {pname if not real_name else real_name} -U",
                    True,
                )
            if not_exit:
                exec((f"from {pname} import {module}" if module else f"import {pname}"), globals(), local_scope)
            else:
                console.print(
                    info_string,
                    f'just run again: "{" ".join(sys.argv)}"'
                    if user_lang != "zh"
                    else f'请重新运行: "{" ".join(sys.argv)}"',
                )
                exit(0)
        else:
            exit(-1)
    finally:
        return local_scope.get(module if module else pname)


from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import numpy

with open("README.md", "r") as fh:
    long_description = fh.read()
VERSION = "0.0.34"

module = Extension(
    "MtxDrawer.MtxReader",
    sources=["MatGen/MatGen.pyx", "MatGen/_MatGen.c"],
    extra_compile_args=["-ffast-math"],
    include_dirs=[numpy.get_include()],
    language="c"
)

setup(
    name='MtxDrawer',
    version=VERSION,
    description='Draw Mtx As Thumbnail',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    author='RhythmLian',
    author_email='RhythmLian@outlook.com',
    url="https://github.com/Rhythmicc/DrawMtxAsThumbnail",
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    ext_modules=cythonize(module),
    install_requires=['numpy', 'matplotlib', 'Qpro', 'cython'],
    entry_points={
        'console_scripts': [
            'mtx-drawer = MtxDrawer.main:main'
        ]
    },
)



from QuickProject.Commander import Commander
from . import console, info_string, status
from .Drawer import Drawer
import os

app = Commander("mtx-drawer")
rt_path = os.path.abspath(os.getcwd())


@app.custom_complete("font_color")
def draw_one():
    return ["black", "white", "red", "green", "blue", "yellow", "cyan", "magenta"]

@app.custom_complete("img_format")
def draw_one():
    return ["svg", "png", "jpg", "jpeg", "bmp", "eps"]

@app.custom_complete("ops")
def draw_one():
    return ['abs', 'log', 'aver', 'real']

@app.custom_complete("color_theme")
def draw_one():
    return ["Default", "SuiteSparse"]

@app.custom_complete('font_color')
def draw():
    return ["black", "white", "red", "green", "blue", "yellow", "cyan", "magenta"]

@app.custom_complete('img_format')
def draw():
    return ["svg", "png", "jpg", "jpeg", "bmp", "eps"]

@app.custom_complete('ops')
def draw():
    return ['abs', 'log', 'aver', 'real']

@app.custom_complete('color_theme')
def draw():
    return ["Default", "SuiteSparse"]

@app.command()
def draw_one(
    filepath: str,
    ops: list,
    force: bool = False,
    log_times: int = 2,
    mat_size: int = 200,
    block_size: int = -1,
    tick_step: int = -1,
    img_format: str = "svg",
    font_color: str = "black",
    color_theme: str = "Default",
    color_bar: bool = False,
    show_in_console: bool = False,
):
    """
    单个文件处理

    :param filepath: 文件路径
    :param ops: 操作列表
    :param force: 是否强制更新
    :param log_times: 取log次数
    :param mat_size: 缩略图尺寸
    :param block_size: 设置块大小（此参数设置后将覆盖mat_size）
    :param img_format: 图片格式
    :param font_color: 字体颜色
    :param show_in_console: 是否在控制台显示图像, 如果为True则不会保存图像
    :return:
    """
    status("处理").start()
    try:
        drawer = Drawer(
            filepath,
            "aver" in ops,
            force,
            log_times,
            set_mat_size=mat_size,
            set_block_size=block_size,
            set_tick_step=tick_step,
            img_format=img_format,
            font_color=font_color,
            set_color_theme=color_theme,
            color_bar=color_bar,
            show_in_console=show_in_console,
        )
    except ValueError:
        return
    for func in ops:
        drawer.call(func)
    status.stop()


@app.command()
def draw(
    ops: list,
    force: bool = False,
    log_times: int = 2,
    mat_size: int = 200,
    block_size: int = -1,
    tick_step: int = -1,
    img_format: str = "svg",
    font_color: str = "black",
    color_bar: bool = False,
    color_theme: str = "Default",
    parallel: bool = False,
):
    """
    多个文件处理
    
    :param ops: 操作列表
    :param force: 是否强制更新
    :param log_times: 取log次数
    :param mat_size: 缩略图尺寸
    :param block_size: 设置块大小（此参数设置后将覆盖mat_size）
    :return:
    """
    if not parallel:
        has_aver = "aver" in ops
        status("遍历并处理").start()
        for rt, _, sonFiles in os.walk(rt_path, followlinks=True):
            for file in sonFiles:
                if file.endswith(".mtx"):
                    try:
                        console.rule(f'{os.path.join(rt, file)}')
                        drawer = Drawer(
                            os.path.join(rt, file),
                            has_aver,
                            force,
                            log_times,
                            set_mat_size=mat_size,
                            set_block_size=block_size,
                            set_tick_step=tick_step,
                            img_format=img_format,
                            font_color=font_color,
                            color_bar=color_bar,
                            set_color_theme=color_theme
                        )
                        for func in ops:
                            drawer.call(func)
                    except ValueError:
                        continue
                    except Exception:
                        console.print_exception()
        status.stop()
        console.print(info_string, "处理完成")
        console.print("-" * console.width)
    else:
        mtx_files = []
        for rt, _, sonFiles in os.walk(rt_path, followlinks=True):
            for file in sonFiles:
                if file.endswith(".mtx"):
                    mtx_files.append(os.path.join(rt, file))

        from concurrent.futures import ThreadPoolExecutor, wait
        
        def draw_one_file(file, progress=None, task_id=None):
            try:
                drawer = Drawer(
                    file,
                    "aver" in ops,
                    force,
                    log_times,
                    set_mat_size=mat_size,
                    set_block_size=block_size,
                    set_tick_step=tick_step,
                    img_format=img_format,
                    font_color=font_color,
                    set_color_theme=color_theme,
                    color_bar=color_bar,
                    show_in_console=False,
                )
                for func in ops:
                    drawer.call(func)
                if progress:
                    progress.advance(task_id)
            except ValueError:
                return
            except Exception:
                return

        from rich.progress import Progress
        progress = Progress()
        task_id = progress.add_task("Drawing", total=len(mtx_files))

        progress.start()
        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(draw_one_file, file, progress, task_id) for file in mtx_files]
            wait(futures)
        progress.stop()


@app.command()
def update():
    """
    更新
    """
    from QuickProject import user_pip, external_exec

    with status("正在更新..."):
        external_exec(f"{user_pip} install -U MtxDrawer")
    console.print(info_string, "更新完成")


def main():
    app()


if __name__ == "__main__":
    main()
